/// @author Wikipedia (original define_MPI_datatype()).
/// @author Andrija Bošnjaković (remove_MPI_datatype(), main()).

#include <mpi.h>
#include <stdio.h>
#include <string.h>

typedef struct
{
	int f;
	short p;
} A;

typedef struct
{
	A a;
	int pp,vp;
} B;

#define MASTER_RANK 0
#define NUM_TYPES 6
#define DUMMY_TAG 42

void define_MPI_datatype(MPI_Datatype* new_type_addr)
{
	int blocklen[NUM_TYPES]={1,1,1,1,1,1}; //The first and last elements mark the beg and end of data structure
	MPI_Aint disp[NUM_TYPES];
	MPI_Datatype type[NUM_TYPES]={MPI_LB, MPI_INT, MPI_SHORT, MPI_INT, MPI_INT, MPI_UB};
	B findsize[2]; //You need an array to establish the upper bound of the data structure
	MPI_Aint findsize_addr, a_addr, f_addr, p_addr, pp_addr, vp_addr, UB_addr;
	int error;

	MPI_Get_address(&findsize[0], &findsize_addr);
	MPI_Get_address(&(findsize[0]).a, &a_addr);
	MPI_Get_address(&((findsize[0]).a).f, &f_addr);
	MPI_Get_address(&((findsize[0]).a).p, &p_addr);
	MPI_Get_address(&(findsize[0]).pp, &pp_addr);
	MPI_Get_address(&(findsize[0]).vp, &vp_addr);
	MPI_Get_address(&findsize[1],&UB_addr);

	disp[0]=a_addr-findsize_addr;
	disp[1]=f_addr-findsize_addr;
	disp[2]=p_addr-findsize_addr;
	disp[3]=pp_addr-findsize_addr;
	disp[4]=vp_addr-findsize_addr;
	disp[5]=UB_addr-findsize_addr;

	error=MPI_Type_create_struct(NUM_TYPES, blocklen, disp, type, new_type_addr);
	MPI_Type_commit(new_type_addr);
}

void remove_MPI_datatype(MPI_Datatype* new_type_addr)
{
	MPI_Type_free(new_type_addr);
}

int main(int argc, char* argv[]) 
{
	int size, rank;
	B dummy;
	MPI_Datatype struct_type;

	memset(&dummy, 0, sizeof dummy);

	MPI_Init(&argc, &argv);
	MPI_Comm_size(MPI_COMM_WORLD, &size);
	MPI_Comm_rank(MPI_COMM_WORLD, &rank);

	define_MPI_datatype(&struct_type);

	if (MASTER_RANK == rank)
	{
		int dest;
		printf("Unesite podatke: \n");
		scanf("%d", &dummy.a.f);
		scanf("%hd", &dummy.a.p);
		scanf("%d", &dummy.pp);
		scanf("%d", &dummy.vp);

		for (dest = 1; dest < size; dest++)
		{
			MPI_Send(&dummy, 1, struct_type, dest, DUMMY_TAG, MPI_COMM_WORLD);
		}
	}
	else
	{
		MPI_Status status;
		MPI_Recv(&dummy, 1, struct_type, MASTER_RANK, DUMMY_TAG, MPI_COMM_WORLD, &status);
		printf("%d:%d\n", rank, dummy.a.f);
		printf("%d:%hd\n", rank, dummy.a.p);
		printf("%d:%d\n", rank, dummy.pp);
		printf("%d:%d\n", rank, dummy.vp);
	}

	remove_MPI_datatype(&struct_type);

	MPI_Finalize();

	return 0;
}
